#--------------------------------
# worddoc.tcl - text extract/overwrite tool for MS-Word file
# Author: Shinsuke Takeuchi <ctrans@gmail.com>
# URL: http://www.karak.jp/tcl/tcl-worddoc.html
# Version 1.0.0
#--------------------------------
package require Tcl 8.5
package require tdom
package provide worddoc 1.0
#--------------------------------
namespace eval worddoc {
	variable stocker ""
	variable content ""
	variable inline ""
	variable delimiter "\ufeff"
	variable markup
	set markup(b) "w:b"
	set markup(i) "w:i"
	set markup(u) "w:u"
	set markup(fg) "w:color"
	set markup(bg) "w:highlight"
	set markup(sz) "w:sz"
	set markup(sub) "w:vertAlign"
	set markup(sup) "w:vertAlign"
	set markup(tbx) "w:txbxContent"
	set markup(fnote) "w:footnote"
	set markup(cbody) "aml:annotation"
	set markup(comment) "aml:content"
	set markup(img) "w:bindata"
	set markup(del) "w:strike"
	set markup(ins) "w:dummy"
	set markup(format) "w:dummy"
	variable wdformat
	set wdformat(doc) 0
	set wdformat(docx) 12
	set wdformat(xml) 11
	set wdformat(rtf) 6
	set wdformat(txt) 2
	variable vbstemplate
	set vbstemplate {\
		Option Explicit\n\
		On Error Resume Next\n\
		Dim objWdApp\n\
		Dim objWdDoc\n\
		Set objWdApp = WScript.CreateObject("Word.Application")\n\
		If Err.Number = 0 Then\n\
			Set objWdDoc = objWdApp.Documents.Open("$src_path")\n\
			If Err.Number = 0 Then\n\
				objWdDoc.SaveAs "$dest_path", $filetype\n\
				objWdDoc.Close\n\
				objWdApp.Quit\n\
			Else\n\
				WScript.Echo Err.Description\n\
			End If\n\
		Else\n\
			WScript.Echo Err.Description\n\
		End If\n\
		Set objWdDoc = Nothing\n\
		Set objWdApp = Nothing\n\
	}
	namespace export extract overwrite changeFormat getInstallationPath
}
#--------------------------------
# worddoc::extract
# extract data from MS-Word file
#--------------------------------
proc worddoc::extract {wordfile {destination {}}} {
	if {![file exists $wordfile]} {return}
	#convert to wordml
	set extension [string tolower [file extension $wordfile]]
	if {$extension != ".xml"} {
		set wordfile [worddoc::changeFormat $wordfile "xml"]
	}
	if {$wordfile == ""} {return}
	#parse xml
	set content [worddoc::getContent $wordfile]
	if {[catch {dom parse $content mydom} err]} {
		error $err
	}
	set root [$mydom documentElement]
	set body [$root getElementsByTagName w:body]
	set worddoc::inline ""
	set worddoc::stocker ""
	set worddoc::content ""
	worddoc::explore $body
	$mydom delete
	#finish explore
	append worddoc::content $worddoc::stocker
	set body ""
	set i 0
	#delete first delimiter
	set worddoc::content [string range $worddoc::content 1 end]
	foreach para [split $worddoc::content $worddoc::delimiter] {
		set para "<$i>$para"
		lappend body $para
		incr i
	}
	#save
	if {$destination == ""} {set destination "[file rootname $wordfile].txt"}
	worddoc::save $destination [join $body \n]
	return $destination
}
#--------------------------------
# worddoc::explore
# ignore inline element(bold, italic, underline, etc.) or not
#--------------------------------
proc worddoc::explore {parent {ignoreInline 0}} {
	variable stocker
	variable content
	variable inline
	set type [$parent nodeType]
	set name [$parent nodeName]
	set pnode [[$parent parentNode] nodeName]
	set block 0 ;#flag for block element
	switch $name {
		\#text -
		cdata {
			if {$pnode == "w:binData" || $pnode == "w:instrText"} {
				set worddoc::inline ""
			} else {
				lassign $worddoc::inline l r
				append worddoc::stocker $l [$parent nodeValue] $r
				set worddoc::inline ""
			}
		}
		v:imagedata {
			append worddoc::stocker "((bindata|img))"
		}
		w:p {
			append worddoc::content $worddoc::stocker
			append worddoc::content "\ufeff"
			set worddoc::stocker ""
		}
		w:r {
			set worddoc::inline ""
		}
		aml:annotation {
			set type [$parent getAttribute w:type]
			set id [$parent getAttribute aml:id]
			if {$type =="Word.Comment.Start"} {
				append worddoc::stocker "(("
			} elseif {$type =="Word.Comment.End"} {
				append worddoc::stocker "|cbody|${id}))"
			} elseif {$type == "Word.Deletion"} {
				append worddoc::stocker "((|del|${id}))"
			} elseif {$type == "Word.Insertion"} {
				append worddoc::stocker "((|ins|${id}))"
			} elseif {$type == "Word.Formatting"} {
				append worddoc::stocker "((|format|${id}))"
			}
		}
		aml:content {
			if {[regexp {(.*)\(\(\|(ins|del)\|([0-9]+)\)\)$} $worddoc::stocker dummy body class id]} {
				append worddoc::content $body
				set worddoc::stocker ""
				append worddoc::stocker "((" [$parent asText] "|$class|${id}))"
				set block 1
			} elseif {[regexp {(.*)\(\(\|(format)\|([0-9]+)\)\)$} $worddoc::stocker dummy body class id]} {
				append worddoc::content $body
				set worddoc::stocker ""
			} else {
				append worddoc::content $worddoc::stocker
				set worddoc::stocker ""
				append worddoc::stocker "((" [$parent asText] "|comment))"
				set block 1
			}
		}
		w:txbxContent {
			append worddoc::content $worddoc::stocker
			set worddoc::stocker ""
			append worddoc::stocker "((" [$parent asText] "|tbx))"
			set block 1
		}
		w:footnote {
			append worddoc::content $worddoc::stocker
			set worddoc::stocker ""
			append worddoc::stocker "((" [$parent asText] "|fnote))"
			set block 1
		}
		w:b {
			if {!$ignoreInline} {
				lassign $worddoc::inline l r
				append l "(("
				set r "|b))$r"
				set worddoc::inline [list $l $r]
			}
		}
		w:u {
			if {!$ignoreInline} {
				lassign $worddoc::inline l r
				append l "(("
				set r "|u))$r"
				set worddoc::inline [list $l $r]
			}
		}
		w:i {
			if {!$ignoreInline} {
				lassign $worddoc::inline l r
				append l "(("
				set r "|i))$r"
				set worddoc::inline [list $l $r]
			}
		}
		w:highlight {
			if {!$ignoreInline} {
				set value [$parent getAttribute w:val]
				lassign $worddoc::inline l r
				append l "(("
				set r "|bg|${value}))${r}"
				set worddoc::inline [list $l $r]
			}
		}
		w:color {
			if {!$ignoreInline} {
				set value [$parent getAttribute w:val]
				lassign $worddoc::inline l r
				append l "(("
				set r "|fg|${value}))${r}"
				set worddoc::inline [list $l $r]
			}
		}
		w:vertAlign {
			if {!$ignoreInline} {
				set type ""
				set value [$parent getAttribute w:val]
				if {$value =="subscript"} {
					set type "sub"
				} elseif {$value =="superscript"} {
					set type "sup"
				}
				if {$type != ""} {
					lassign $worddoc::inline l r
					append l "(("
					set r "|${type}))$r"
					set worddoc::inline [list $l $r]
				}
			}
		}
	}
	if {!$block} {
		foreach child [$parent childNodes] {
			worddoc::explore $child
		}
	}
}
#--------------------------------
# worddoc::overwrite
# overwrite ms-word file($originalFile) with content of $textFile, will return filename(destination) or empty
#--------------------------------
proc worddoc::overwrite {textFile originalFile {destination {}}} {
	set errorLog [list]
	set content [worddoc::getContent $textFile]
	regsub -all {\n([^<])} $content {\1} content
	set content [split $content \n]
	#convert to wordml
	if {[file extension $originalFile] != ".xml"} {
		set originalFile [worddoc::changeFormat $originalFile xml]
	}
	if {![file exists $originalFile]} {
		error "Can't open $originalFile"
	}
	set xml [worddoc::getContent $originalFile]
	if {[catch {dom parse $xml mydom} err]} {
		error "Parse Error"
	}
	set root [$mydom documentElement]
	set body [$root getElementsByTagName w:body]
	set i 0
	#replace all element of tree
	foreach para [$body descendant all w:p] {
		set worddoc::stocker [list]
		foreach run [$para descendant all w:r] {
			#store data of image, textbox, footnote
			foreach child [$run childNodes] {
				set cnode [$child nodeName]
				switch $cnode {
					aml:annotation -
					w:pict -
					w:footnote {
						lappend worddoc::stocker [$child asXML]
					}
				}
			}
			$run delete
		}
		set str [lindex $content $i]
		#check line-number
		if {[regexp -- {^<([0-9]+)>} $str match num]} {
			if {[expr $num - $i] == 0} {
				set str [string range $str [string length $match] end]
				worddoc::assign $para $str
			} else {
				lappend errorLog "NUMBER: $str"
			}
		} else {
			lappend errorLog "LINEHEAD: $str (No.$i)"
		}
		incr i
	}
	#error check
	if {[llength $errorLog] != 0} {
		$mydom delete
		set fn "[file rootname $originalFile].log"
		worddoc::save $fn [join $errorLog \n]
		error "Some errors occurred. See the error log for more details."
	}
	#destination
	if {$destination == ""} {
		set destination "[file rootname $originalFile]-out.doc"
	}
	if {[file extension $destination] != ".xml"} {
		set temp [file join [file dirname $destination] temporary.xml]
		worddoc::save $temp [$root asXML]
		set format [string range [file extension $destination] 1 end]
		worddoc::changeFormat $temp $format $destination
		if {![file exists $destination]} {
			$mydom delete
			error "Can't open $destination"
		}
		after idle {catch {file delete -force $temp}}
	} else {
		worddoc::save $destination [$root asXML]
	}
	$mydom delete
	return $destination
}
#--------------------------------
#	worddoc::assign
#	apply $str to $parent
#--------------------------------
proc worddoc::assign {parent str} {
	set pos 0
	set start 0
	set openTag "(("
	set closeTag "))"
	set nest 0
	set body ""
	set tags [list]
	while {1} {
		set current [string range $str $pos [expr $pos + 1]]
		if {[string length $current] <= 1} {break}
		#start new tag
		if {$current == $openTag} {
			if {$nest == 0} {
				worddoc::push $parent [string range $str $start [expr $pos - 1]]
			}
			incr nest
			incr pos 2
			set start $pos
		} elseif {$current == $closeTag} {
			set content [string range $str $start [expr $pos - 1]]
			lassign [split $content |] data tag value
			incr nest -1
			if {$nest < 0} {
				#open tag and cloase tag are not match
				set nest 0
			} elseif {$nest == 0} {
				#end of current tag
				if {[info exists worddoc::markup($tag)]} {
					append body $data
					lappend tags [list $tag $value]
				} else {
					append body $openTag $data $closeTag
				}
				worddoc::push $parent $body $tags
				set body ""
				set tags [list]
			} elseif {$nest > 0} {
				if {[info exists worddoc::markup($tag)]} {
					append body $data
					lappend tags [list $tag $value]
				} else {
					append body $openTag $data $closeTag
				}
			}
			incr pos 2
			set start $pos
		} else {
			#normal text
			incr pos
		}
	}
	#remaining data
	worddoc::push $parent [string range $str $start end] $tags
}
#--------------------------------
#	worddoc::push
#	append $tag and $str to $parent
#--------------------------------
proc worddoc::push {parent str {tags {}}} {
	variable markup
	variable stocker
	set id ""
	set doc [dom createDocument w:r]
	set root [$doc documentElement]
	#書式（w:rPr）の設定
	if {[llength $tags] > 0} {
		set node [$doc createElement w:rPr]
		$root appendChild $node
		foreach tag $tags {
			lassign $tag tagname value
			if {$tagname == ""} {continue}
			if {![info exists worddoc::markup($tagname)]} {
				set subnode [$doc createElement w:color]
				$node appendChild $subnode
				$subnode setAttribute w:val "red"
				set str "error: $str"
				continue
			}
			set name $worddoc::markup($tagname)
			switch $name {
				w:b -
				w:i -
				w:color -
				w:highlight -
				w:sz {
					set subnode [$doc createElement $name]
					$node appendChild $subnode
					if {$value != ""} {$subnode setAttribute w:val $value}
				}
				w:u  {
					set value "single"
					set subnode [$doc createElement $name]
					$node appendChild $subnode
					$subnode setAttribute w:val $value
				}
				w:vertAlign {
					set value "superscript"
					if {$tagname == "sub"} {set value "subscript"}
					set subnode [$doc createElement $name]
					$node appendChild $subnode
					$subnode setAttribute w:val $value
				}
				w:txbxContent -
				w:footnote -
				aml:content {
					worddoc::replaceElement $root $str
					set str ""
				}
				w:strike {
					set subnode [$doc createElement $name]
					$node appendChild $subnode
				}
				aml:annotation {
					#親ノードを検索してコメントの開始と終了タグを削除、ループ終了後に新しくノードを挿入
					foreach child [$parent childNodes] {
						if {$name == [$child nodeName]} {
							set type [$child getAttribute w:type]
							if {$type == "Word.Comment.Start" || $type == "Word.Comment.End"} {
								set id $value
								$child delete
							}
						}
					}
				}
				w:bindata {
					set xml [lindex $worddoc::stocker 0]
					set worddoc::stocker [lrange $worddoc::stocker 1 end]
					if {$xml != ""} {$node appendXML $xml}
					set str ""
				}
			}
		}
	}
	#文字列をノードに追加
	if {$str != ""} {
		set node [$doc createElement w:t]
		$node appendChild [$doc createTextNode $str]
		$root appendChild $node
	}
	$parent appendChild $root
	if {$id != ""} {
		set aml [$doc createElement aml:annotation]
		$parent insertBefore $aml $root
		$aml setAttribute aml:id $id w:type "Word.Comment.Start"
		set aml [$doc createElement aml:annotation]
		$parent appendChild $aml
		$aml setAttribute aml:id $id w:type "Word.Comment.End"
	}
}
#--------------------------------
#	worddoc::replaceElement (for footnote & textbox)
#	delete text element of $node and insert $str
#--------------------------------
proc worddoc::replaceElement {node str} {
	variable stocker
	set xml [lindex $worddoc::stocker 0]
	set worddoc::stocker [lrange $worddoc::stocker 1 end]
	if {$xml == ""} {return}
	dom parse $xml doc
	set child [$doc documentElement]
	set textNodes [$child getElementsByTagName w:t]
	set textNode [lindex $textNodes 0]
	if {$textNode == ""} {return}
	set pnode [$textNode parentNode]
	foreach nd $textNodes {
		if {$nd != ""} {$nd delete}
	}
	set new [$doc createElement w:t]
	$new appendChild [$doc createTextNode $str]
	$pnode appendChild $new
	$node appendChild $child
}
#--------------------------------
# worddoc::changeFormat
#--------------------------------
proc worddoc::changeFormat {wordfile format {destination {}}} {
	global env
	#source-file
	if {![file exists $wordfile]} {return}
	set src_path [file nativename $wordfile]
	#destination
	set format [string tolower $format]
	if {$destination == ""} {
		set rootname [file rootname $wordfile]
		set dest_path [file nativename [append rootname "." $format]]
	} else {
		set dest_path $destination
	}
	if {[file exists $destination]} {
		catch {file delete -force $destination}
	}
	#constant of file-format
	set filetype $worddoc::wdformat($format)
	#vbs template
	set template [subst -nocommands $worddoc::vbstemplate]
	set vbsf [file join $env(HOME) "word.vbs"]
	worddoc::save $vbsf $template [encoding system]
	#execute vbs
	exec wscript $vbsf &
	set i 0
	while {1} {
		incr i
		after 10
		if {[file exists $dest_path]} {break}
		if {$i >= 1000} {return}
	}
	return $dest_path
}
#--------------------------------
# worddoc::getInstallationPath
#--------------------------------
proc worddoc::getInstallationPath {} {
	package require registry
	return [registry get {HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\Winword.exe} {}]
}
#--------------------------------
#	worddoc::save
#--------------------------------
proc worddoc::save {fn dat {enc utf-8} {nl auto}} {
	set f [open $fn w]
	fconfigure $f -encoding $enc -translation $nl
	if {$enc == "unicode"} {puts -nonewline $f "\ufeff"}
	puts $f $dat
	close $f
}
#--------------------------------
#	worddoc::getContent
#--------------------------------
proc worddoc::getContent {fn {enc utf-8} {nl auto}} {
	if {![file exists $fn]} {return}
	set f [open $fn r]
	fconfigure $f -encoding $enc -translation $nl
	set body [read $f]
	close $f
	if {[string index $body 0] == "\ufeff"} {
		set body [string range $body 1 end]
	}
	return $body
}